#include <sys/time.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>

#include <semaphore.h>

#include <iostream>
#include <set>
#include <sstream>
#include <cstring>
#include <cstdlib>
#include <fstream>
#include <cstdio>
#include <map>
#include <queue>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <vector>

using namespace std;

int server_sock;
ofstream ServerLog;
string fileLogServer;
sem_t mutexQueue;

struct Request {
	int idClienta;
	int row;
};

struct ChangeRows {
	int row;
	int value;
	int timestamp;
	timeval receiveTime;
};

struct Response {
	int row;
	int value;
	int timestamp;
	timeval sendTime;
	timeval receiveTime;
};

struct Row {
	int id_row;
	int value;
	int timestamp;
};
Row database[100];

struct ResultCommit {
	bool commit;
	timeval timeCommit;
};

timeval getTime() {
	timeval time;
	gettimeofday(&time, NULL);
	return time;
}

string timeToString(timeval cas) {
	stringstream ss;
	ss << cas.tv_sec << "." << cas.tv_usec;
	return ss.str();
}

string numberToString(int number) {
	stringstream ss;
	ss << number;
	return ss.str();
}

void writeToLogFromClient(timeval casOdeslani, timeval casPrijeti, string data,
		int idClient) {
	ServerLog.open(fileLogServer.c_str(), std::ios_base::app);
	ServerLog << "- dst: {p: SERVER, ts: " << timeToString(casPrijeti) << "}\n";
	ServerLog << "  src: {p: P" << numberToString(idClient) << ", ts: "
			<< timeToString(casOdeslani) << "}\n";
	ServerLog << "  data: '" << data << "'\n";
	ServerLog.close();
}

void writeToLog(timeval sendTime, timeval receiveTime, string data,
		int idClient) {
	ServerLog.open(fileLogServer.c_str(), std::ios_base::app);
	ServerLog << "- src: {p: SERVER, ts: " << timeToString(sendTime) << "}\n";
	ServerLog << "  dst: {p: P" << numberToString(idClient) << ", ts: "
			<< timeToString(receiveTime) << "}\n";
	if (data.compare("commit") == 0) {
		ServerLog << "  color: '#00FF00'\n";
		ServerLog << "  data: 'COMMIT'\n";
	} else if (data.compare("abort") == 0) {
		ServerLog << "  color: '#FF0000'\n";
		ServerLog << "  data: 'ABORT'\n";
	} else {
		ServerLog << "  data: '" << data << "'\n";
	}
	ServerLog.close();
}

void settingServer(int port) {
	struct sockaddr_in local_addr;

	server_sock = socket(AF_INET, SOCK_STREAM, 0);
	if (server_sock < 0) {
		cout << "The problem occurred during creating socket.\n";
		exit(EXIT_FAILURE);
	}

	int reuse = 1;
	setsockopt(server_sock, SOL_SOCKET, SO_REUSEADDR, (char *) &reuse,
			sizeof(reuse));

	memset(&local_addr, 0, sizeof(struct sockaddr_in));

	local_addr.sin_family = AF_INET;
	local_addr.sin_port = htons(port);
	local_addr.sin_addr.s_addr = INADDR_ANY;

	if (bind(server_sock, (struct sockaddr *) &local_addr,
			sizeof(struct sockaddr_in)) != 0) {
		cout << "The problem occurred during binding.\n";
		exit(EXIT_FAILURE);
	}

	if (listen(server_sock, 5) != 0) {
		cout << "The problem occurred during listen.\n";
		exit(EXIT_FAILURE);
	}

	cout << "Server has been started...\n";
}

void createDatabase() {
	for (int i = 0; i < 100; i++) {
		database[i].id_row = i;
		database[i].value = -50 + i;
		database[i].timestamp = 0;
	}
}

void *listenClient(void *arg) {
	set<int> requestRows;
	int idClient;
	int client_sock;
	client_sock = *(int *) arg;

	Request request;
	timeval timeReceive;
	timeval receiveBegin;
	bool theFirst = true;
	while (true) {
		read(client_sock, (void *) &request, sizeof(Request));

		if (theFirst) {
			receiveBegin = getTime();
			theFirst = false;
		}

		if (request.idClienta == -1)
			break;

		idClient = request.idClienta;
		timeReceive = getTime();
		requestRows.insert(request.row);
	}

	sleep(1);
	writeToLogFromClient(receiveBegin, getTime(), "BEGIN", idClient);

	/**
	 * Prijmul jsem pozadavky poslu radky zpet co chce...
	 */

	for (set<int>::iterator i = requestRows.begin(); i != requestRows.end();
			i++) {
		int row = *i;
		Response tmp;
		tmp.receiveTime = getTime();
		tmp.sendTime = getTime();
//		sem_wait(&mutexQueue);
		tmp.timestamp = database[row].timestamp;
		tmp.value = database[row].value;
		tmp.row = database[row].id_row;
//		sem_post(&mutexQueue);
		write(client_sock, (void *) &tmp, sizeof(Response));
	}
	Response endRes;
	endRes.row = -1;
	write(client_sock, (void *) &endRes, sizeof(Response));

	/**
	 * radky odeslany, prijmu upravene radky
	 */

	vector<ChangeRows> changeVector;
	ChangeRows row;
	//timeval timeCommit;
	timeval receiveTime;
	while (true) {
		read(client_sock, (void *) &row, sizeof(ChangeRows));
		if (row.row == -1)
			break;
		//timeCommit = getTime();
		receiveTime = row.receiveTime;
		changeVector.push_back(row);
	}
	sleep(1);
	writeToLogFromClient(receiveTime, getTime(), "NEW_DATA", idClient);

	bool readyToCommit = true;
	//kontrola udaju
//	sem_wait(&mutexQueue);
	for (unsigned int i = 0; i < changeVector.size(); i++) {
		if (changeVector[i].timestamp
				!= database[changeVector[i].row].timestamp) {
			readyToCommit = false;
		}
	}
//	sem_post(&mutexQueue);

	ResultCommit vc;
	vc.timeCommit = getTime();
	if (readyToCommit == true) {
//		sem_wait(&mutexQueue);
		for (unsigned int i = 0; i < changeVector.size(); i++) {
			database[changeVector[i].row].value = changeVector[i].value;
			database[changeVector[i].row].timestamp++;
		}
//		sem_post(&mutexQueue);
		vc.commit = true;
	} else {
		vc.commit = false;
	}

	write(client_sock, (void *) &vc, sizeof(ResultCommit));

//	string result = "abort";
//	if (vc.commit == true) {
//		result = "commit";
//	}
//
//	writeToLog(getTime(), getTime(), result, idClient);

	return NULL;
}

int main(int argc, char* argv[]) {

	if (argc != 2) {
		cout << "./Server <port>\n";
		exit(1);
	}

	settingServer(atoi(argv[1]));
	createDatabase();
	fileLogServer = "../log/log_server.txt";
	//sem_init(&mutexQueue, 0, 1);
	int client_sock;
	struct sockaddr_in remote_addr;
	socklen_t remote_addr_len;
	void *th_socket;

	while (true) {
		pthread_t workThread;
		client_sock = accept(server_sock, (struct sockaddr *) &remote_addr,
				&remote_addr_len);
		if (client_sock > 0) {
			th_socket = malloc(sizeof(int));
			th_socket = &client_sock;
			pthread_create(&workThread, NULL, listenClient, (void*) th_socket);
		}
	}

	return 0;
}
